from pbu.default_options import list_join


class DebugObject:
    def __init__(self, debug=False, logger=None):
        self._debug = debug
        self._logger = logger

    def debug(self, *kwargs):
        if self._debug:
            if self._logger is not None:
                self._logger.info(list_join(kwargs, " "))
            else:
                print(*kwargs)



import os
import json
from typing import Optional, Union
from pbu.default_options import default_options


def write_json(data: Union[dict, list], path: str):
    """
    Writes an object to a json file. This function will take care of opening and closing the file.
    :param data: a list of dictionary - has to be possible to serialise it as JSON.
    :param path: the file path where to write the file to
    """
    if data is None or not isinstance(data, (list, dict)):
        raise ValueError("No or invalid data provided")
    fp = open(path, "w")
    json.dump(data, fp)
    fp.close()


def read_json(path: str) -> Optional[Union[dict, list]]:
    """
    Reads from a json file if it exists and returns the content. This function will take care of opening and closing the
    file.
    :param path: the file path of the json file
    :return: the list of dictionary contained in the json file
    """
    if not os.path.exists(path):
        return None
    fp = open(path)
    data = json.load(fp)
    fp.close()
    return data


def ensure_directory(path: str):
    """
    Makes sure a certain directory exists. If it doesn't exist, the directory will be created.
    :param path: a path reference (absolute or relative) to the directory that should exist.
    """
    if not os.path.isdir(path):
        os.makedirs(path)


def convert_to_path(identifier: Optional[str], custom_replacements={}) -> Optional[str]:
    if identifier is None:
        return None

    default_replacements = {
        "-": [" ", "(", ")", "|", "[", "]", ".", ",", "/", "\\"],  # replace these values with hyphen
        "": ["`", '"', "'"]  # replace these values with empty string
    }

    replacement_map = {}
    for replacement, searches in default_replacements.items():
        for search in searches:
            replacement_map[search] = replacement

    if custom_replacements is not None:
        replacement_map = default_options(replacement_map, custom_replacements)

    for search, replace in replacement_map:
        identifier = identifier.replace(search, replace)

    return identifier






from typing import List, Union, Optional, Tuple
from statistics import mean


def weighted_mean(values: List[Union[float, int]], weights: List[Union[int, float]] = []) -> Optional[float]:
    """
    This will generate a mean value for the list of provided `values`, where each value is multiplied by the
    corresponding weight in the same position. If there are more `values` than `weights`, remaining values will receive
    weight 1.0. If there's more `weights` than `values`, it will only use the first `n` weights, where `n` is the number
    of values.
    :param values: the numeric values to generated a weighted mean over
    :param weights: the weights for each position of the values
    :return: the weighted average value of the provided values
    """
    if len(values) == 0:
        return None

    if weights is None or len(weights) == 0:
        return mean(values)

    total_weight = 0.0
    total_value = 0.0

    num_weights = len(weights)

    for idx, val in enumerate(values):
        current_weight = weights[-1] if num_weights < idx + 1 else weights[idx]
        total_weight += current_weight
        total_value += val * current_weight

    if total_weight == 0.0:
        return None

    return total_value / total_weight


def normalise(value: Union[float, int], min_val: Union[float, int], max_val: Union[float, int], limit=True,
              mid_point: Optional[Union[float, int]] = None) -> float:
    """
    Normalises the given input `value` between min_val and max_val as number between 0 and 1. If the min_val is provided
    larger than the max_val, the function will automatically invert them and provide the `1 - normalised_value`.
    :param value: the number to normalise
    :param min_val: the lower boundary representing 0.0
    :param max_val: the upper boundary representing 1.0
    :param limit: a boolean flag indicating whether a value between 0 and 1 is returned. Otherwise, values greater than
    1 and lower than 0 can be returned.
    :param mid_point: an optional mid-point to use for the normalisation. If provided, the normalisation will be
    performed between the min_val and the mid_point, and the mid_point and the max_val.
    :return: the normalised value expressed as where the value sits between min_val and max_val
    """
    if value is None or min_val is None or max_val is None:
        return 0.0  # invalid input

    min_value = min_val
    max_value = max_val
    inverted = False
    if max_val < min_val:
        inverted = True
        min_value = max_val
        max_value = min_val

    if value < min_value and limit is True:
        return 0.0 if inverted is False else 1.0

    if value > max_value and limit is True:
        return 1.0 if inverted is False else 0.0

    if max_value == min_value:
        return 0.5

    def _handle_return(norm):
        return norm if inverted is False else 1.0 - norm  # handle inverted flag

    if mid_point is None:
        return _handle_return((float(value) - float(min_value)) / (float(max_value) - float(min_value)))
    else:
        if value <= mid_point:
            # value is below mid-point (return 0 - 0.5)
            return _handle_return(((float(value) - float(min_value)) / (float(mid_point) - float(min_value))) * 0.5)
        else:
            # value is above mid-point (return 0.5 - 1.0)
            return _handle_return(
                (((float(value) - float(mid_point)) / (float(max_value) - float(mid_point))) * 0.5) + 0.5)


def discretise(value: Union[float, int], precision: Union[float, int] = 1, floor=False,
               ceil=False) -> Union[float, int]:
    """
    This function will round the given value to the nearest multiple of the given precision. There are 2 boolean flags
    available that can force the function to return the lower or upper boundary of a precision interval. If neither is
    set, the function will return the mid-point of the interval.
    :param value: the value to discretise
    :param precision: the precision to use for the discretisation
    :param floor: a boolean flag indicating whether the lower boundary of the precision interval should be returned
    :param ceil: a boolean flag indicating whether the upper boundary of the precision interval should be returned
    :return: the discretised value
    """
    if value is None:
        return None
    remainder = value % precision
    if ceil is True:
        return value - remainder + precision
    if floor is True:
        return value - remainder
    # mid point is default fallback
    return (value - remainder) + (precision * 0.5)


def compute_linear_function_parameters(xy_points: List[tuple]) -> Tuple[float, float, float]:
    """
    Computes m and b to minimise the error of all given points to map onto a linear function y = m * x + b.
    :param xy_points: a list of tuples representing the (x, y) points
    :return: a tuple with 3 elements containing the parameters m, b and the total (sum) error of all points when
    comparing against the linear function.
    """
    n = len(xy_points)

    # arrays to store all the elements
    x2 = []  # x squared
    xy = []  # x * y
    x = []  # all x-values
    y = []  # all y-values

    # collect all point data
    for (x_val, y_val) in xy_points:
        x2.append(x_val * x_val)
        xy.append(x_val * y_val)
        x.append(x_val)
        y.append(y_val)

    # determine m and b
    m = ((n * sum(xy)) - (sum(x) * sum(y))) / ((n * sum(x2)) - (sum(x) * sum(x)))
    b = (sum(y) - (m * sum(x))) / n

    # calculate total error
    error = 0.0
    for (x, y) in xy_points:
        error += abs(y - ((m * x) + b))

    return m, b, error



import warnings
from datetime import datetime, date, time
from abc import ABC, abstractmethod
from typing import Optional, List, Dict


def _handle_value_to_string(value, custom_mapping=None):
    if value is None or isinstance(value, (str, int, float, bool)):
        # primitive type
        return value
    elif isinstance(value, JsonDocument):
        # value is another json document
        return value.to_json()
    elif isinstance(value, list):
        if len(value) == 0:
            # empty list
            return []
        # can contain any supported value
        return list(map(lambda val: _handle_value_to_string(val, custom_mapping), value))
    elif isinstance(value, dict):
        # handle non-string keys
        result = {}
        for k, v in value.items():
            result[str(k)] = _handle_value_to_string(v)
        return result
    elif isinstance(value, (datetime, date, time)):
        if custom_mapping is not None and isinstance(custom_mapping, str):
            return value.strftime(custom_mapping)
    return str(value)  # attempt to convert value to a string


def _parse_json_value(value, custom_mapping=None):
    if value is None or isinstance(value, (float, bool)):
        # primitive type
        return value
    if isinstance(value, (str, int)):
        # could be datetime
        if custom_mapping is not None:
            if isinstance(custom_mapping, str):
                # simply a datetime string (default behaviour)
                return datetime.strptime(value, custom_mapping)
            if isinstance(custom_mapping, tuple):
                # first token is the type, second token the format
                return custom_mapping[0].strptime(value, custom_mapping[1])
            if callable(custom_mapping):
                # custom mapping is a function
                return custom_mapping(value)
        return value
    elif isinstance(value, list):
        # can contain any supported value
        return list(map(lambda val: _parse_json_value(val, custom_mapping), value))
    elif isinstance(value, dict):
        if isinstance(custom_mapping, type) and issubclass(custom_mapping, JsonDocument):
            # custom mapping is a type, attempt to convert dictionary into instance of that type
            return custom_mapping.from_json(value)
        result = {}
        for k, v in value.items():
            result[k] = _parse_json_value(v, custom_mapping)
        return result
    return value


class JsonDocument(ABC):
    def __init__(self):
        pass

    def extract_system_fields(self, json: dict):
        """
        Extracts the id and version from a JSON object or dictionary and maps them to the current instances attributes.
        :param json: the json object or dictionary from which to extract information.
        """
        # check if the get_attribute_mapping method is overridden
        attr_mapping = self._get_attribute_mapping()
        dt_mapping = self._get_custom_mapping()
        if attr_mapping is None:
            return

        # evaluate attribute mapping
        for key in attr_mapping:
            if attr_mapping[key] in json:
                json_val = json[attr_mapping[key]]
                self.__setattr__(key, _parse_json_value(json_val, dt_mapping.get(key, None)))

    def to_json(self) -> dict:
        """
        Returns a serializable representation of this document as dictionary or JSON object.
        :return: a dictionary or JSON object providing the data contained within this document
        """
        result = {}
        attr_mapping = self._get_attribute_mapping()
        custom_mapping = self._get_custom_mapping()
        if attr_mapping is not None:
            for key in attr_mapping:
                if self.__getattribute__(key) is not None:
                    # jsonify value
                    result[attr_mapping[key]] = _handle_value_to_string(self.__getattribute__(key),
                                                                        custom_mapping.get(key, None))

        return result

    @classmethod
    def from_json(cls, json: dict):
        """
        Receives a dictionary or JSON object and returns an instance of this JsonDocument subclass.
        :param json: a dictionary or JSON object instance
        :return: an instance of a subclass of JsonDocument
        """
        obj = cls()
        obj.extract_system_fields(json)
        return obj

    def get_attribute_mapping(self) -> dict:
        """
        Provides a mapping from internal attribute names to JSON attribute names.
        """
        pass

    def get_custom_mapping(self) -> dict:
        """
        Provides a list of internal attributes that represent datetime attributes or custom JsonDocuments, and the value
        is the datetime type (date, datetime, time) and format string provided as tuple or the type that it should be
        parsed into.
        """
        pass

    def _get_attribute_mapping(self) -> Optional[dict]:
        """
        Internal method used to find out if the subclass defines an attribute mapping. If the subclass defines an
        attribute mapping and returns a dictionary, the attribute mapping will be returned. Otherwise, None will be
        returned, which will be used by the to_json and extract_system_fields method to map all primitive fields from
        the de-serialised class to JSON and back.
        """
        # find out if the subclass defines the method
        defining_class = self.get_attribute_mapping.__func__.__qualname__.split(".")[0]
        if defining_class == "JsonDocument":
            return None

        # check if the subclass method returns a dictionary
        attr_mapping = self.get_attribute_mapping()
        if not isinstance(attr_mapping, dict):
            return None

        # mapping provided by subclass, return it
        return attr_mapping

    def _get_custom_mapping(self) -> Optional[dict]:
        """
        Internal method used to find out if the subclass defines a datetime mapping. If the subclass defines a
        datetime mapping and returns a dictionary, the datetime mapping will be returned. Otherwise, an empty dict will
        be returned. This will be used by the to_json and extract_system_fields method to map fields containing datetime
        objects to string and back.
        """
        # find out if the subclass defines the method
        defining_class = self.get_custom_mapping.__func__.__qualname__.split(".")[0]
        if defining_class == "JsonDocument":
            return {}

        # check if the subclass method returns a dictionary
        attr_mapping = self.get_custom_mapping()
        if not isinstance(attr_mapping, dict):
            return {}

        # mapping provided by subclass, return it
        return attr_mapping

    def apply_updates(self, update, attributes: List[str] = []):
        """
        Applies an update (which has to be of the same type as self) to the current instance. The list of attributes
        past will be checked, if they are available.
        :param update: an instance of the same type as self
        :param attributes: a list of strings representing the attributes to update (see get_attribute_mapping)
        :raises ValueError: in case the provided `update` is an instance of a different class than self.
        """
        if not isinstance(update, type(self)):
            raise ValueError("Provided `update` parameter has a different type than `self`")
        if update is None:
            return
        for attr in attributes:
            if hasattr(self, attr) and hasattr(update, attr):
                self.__setattr__(attr, getattr(update, attr))
            else:
                warnings.warn(f"Trying to update attribute '{attr}' but either the item or the update does not have it")


def list_to_json(item_list: List[JsonDocument]):
    """
    Helper class serialising a list of JsonDocument into a list of dictionaries that can easily be serialised.
    :param item_list: a list of JsonDocument instances
    :return: a list of dictionaries.
    """
    return list(map(lambda item: item.to_json(), item_list))


def list_from_json(json_list: List[dict], deserialize_class: JsonDocument):
    """
    Helper class deserialising a list of dictionaries into a list of JsonDocument instances.
    :param json_list: a list of dictionaries
    :param deserialize_class: the class to use for deserialisation
    :return: a list of JsonDocument instances
    """
    if not issubclass(deserialize_class, JsonDocument):
        raise ValueError("Provided `deserialize_class` is not a subclass of JsonDocument")
    return list(map(lambda json: deserialize_class.from_json(json), json_list))
